@@ -272,7 +272,7 @@ fhandler_process::open (int flags, mode_t mode)
 	}
       else
 	{
-	  flags |= O_DIROPEN;
+	  diropen = true;
 	  goto success;
 	}
     }
@@ -287,7 +287,7 @@ fhandler_process::open (int flags, mode_t mode)
     }
   if (entry->fhandler == FH_PROCESSFD)
     {
-      flags |= O_DIROPEN;
+      diropen = true;
       goto success;
     }
   if (flags & O_WRONLY)
@@ -398,6 +398,21 @@ format_process_fd (void *data, char *&destbuf)
 	*((process_fd_t *) data)->fd_type = virt_fdsymlink;
       else /* trailing path */
 	{
+	  /* Does the descriptor link point to a directory? */
+	  bool dir;
+	  if (*destbuf != '/')	/* e.g., "pipe:[" or "socket:[" */
+	    dir = false;
+	  else
+	    {
+	      path_conv pc (destbuf);
+	      dir = pc.isdir ();
+	    }
+	  if (!dir)
+	    {
+	      set_errno (ENOTDIR);
+	      cfree (destbuf);
+	      return -1;
+	    }
 	  char *newbuf = (char *) cmalloc_abort (HEAP_STR, strlen (destbuf)
 							   + strlen (e) + 1);
 	  stpcpy (stpcpy (newbuf, destbuf), e);
@@ -1242,9 +1257,9 @@ format_process_status (void *data, char *&destbuf)
 				   "VmStk:\t%8lu kB\n"
 				   "VmExe:\t%8lu kB\n"
 				   "VmLib:\t%8lu kB\n"
-				   "SigPnd:\t%016x\n"
-				   "SigBlk:\t%016x\n"
-				   "SigIgn:\t%016x\n",
+				   "SigPnd:\t%016lx\n"
+				   "SigBlk:\t%016lx\n"
+				   "SigIgn:\t%016lx\n",
 			  cmd,
 			  state, state_str,
 			  p->pgid,
