@@ -811,7 +811,8 @@ get_posix_access (PSECURITY_DESCRIPTOR psd,
 			  class_perm = lacl[pos].a_perm;
 			}
 		    }
-		  if (ace->Header.AceFlags & SUB_CONTAINERS_AND_OBJECTS_INHERIT)
+		  if (S_ISDIR (attr)
+		      && (ace->Header.AceFlags & SUB_CONTAINERS_AND_OBJECTS_INHERIT) != 0)
 		    {
 		      if ((pos = searchace (lacl, MAX_ACL_ENTRIES,
 					    DEF_CLASS_OBJ)) >= 0)
@@ -849,14 +850,12 @@ get_posix_access (PSECURITY_DESCRIPTOR psd,
       else if (ace_sid == well_known_creator_owner_sid)
 	{
 	  type = DEF_USER_OBJ;
-	  types_def |= type;
 	  id = ACL_UNDEFINED_ID;
 	  saw_def_user_obj = true;
 	}
       else if (ace_sid == well_known_creator_group_sid)
 	{
 	  type = DEF_GROUP_OBJ;
-	  types_def |= type;
 	  id = ACL_UNDEFINED_ID;
 	  saw_def_group_obj = true;
 	}
@@ -952,7 +951,8 @@ get_posix_access (PSECURITY_DESCRIPTOR psd,
 		standard_ACEs_only = false;
 	    }
 	}
-      if ((ace->Header.AceFlags & SUB_CONTAINERS_AND_OBJECTS_INHERIT))
+      if (S_ISDIR (attr)
+	  && (ace->Header.AceFlags & SUB_CONTAINERS_AND_OBJECTS_INHERIT) != 0)
 	{
 	  if (type == USER_OBJ)
 	    {
@@ -1197,10 +1197,11 @@ int
 getacl (HANDLE handle, path_conv &pc, int nentries, aclent_t *aclbufp)
 {
   security_descriptor sd;
+  mode_t attr = pc.isdir () ? S_IFDIR : 0;
 
   if (get_file_sd (handle, pc, sd, false))
     return -1;
-  int pos = get_posix_access (sd, NULL, NULL, NULL, aclbufp, nentries);
+  int pos = get_posix_access (sd, &attr, NULL, NULL, aclbufp, nentries);
   syscall_printf ("%R = getacl(%S)", pos, pc.get_nt_native_path ());
   return pos;
 }
